'''this module is used to do something with fasta
Dev: 
1) parse plastid genome and GFF3 file annotated from GeSeq and concatenate CDS sequences together into a whole-length gene sequence 
Only if the gene has multiple CDS, or it will just save one CDS sequence.
2) DNA sequence translation
'''
def ParseFasta(filename):
    '''Parse the fasta file with header , generated by TBtools.
    TBtools:
    java -cp <Path_to_TBtools>/TBtools_JRE1.6.jar biocjava.bioIO.GFF.ExtractFeaturefromGFF3andGenome \
    --inGtf Geseq.gff3 \
    --inGenome plastid.assembly \
    --outFile <output>.fasta \
    --targetFeature <selected_features, e.g. > --targetIdTag ID --retainAttr true --onlyCheck false --maxFeatureCounts 10 --minFeatureCounts 1 --upStreamBases 0 --downStreamBases 0 --onlyRetainFlank false --addN false
    '''
    # with open(options.Geseq_out, "r") as f:
    with open(filename, "r") as f:
        data = f.readlines()
        chr_gene_pos = {}
        gene_seq = {}
        chr_list = []
        for line in data:
            if line.startswith(">"):
                seq_header = line.strip()
                seq_info = line.split()
                for i in seq_info:
                    if "Location" in i:
                        pos = i
                        seqchr = pos.split("=")[1].split(":")[0].strip()
                        chr_list.append(seqchr)
                        seqstart = pos.split(":")[1]
                        chr_gene_pos.setdefault(seqchr, {})
                chr_gene_pos[seqchr][seq_header] = seqstart
                gene_seq[seq_header] = []
            else:
                gene_seq[seq_header].append(line.strip())
    return chr_gene_pos, gene_seq, chr_list

def SortByChromosome(chr_gene_pos, gene_seq, chr_list, outname, header_len):
    '''Sort the retrieved sequence based on their chr_number and start position'''
    chr_list = set(chr_list)
    if len(chr_list) > 1:
        sort_seq_chr = sorted(chr_list, key = lambda d:int(d[header_len:]), reverse = False)
    else:
        sort_seq_chr = chr_list
    # with open(options.outname, "w") as out:
    with open(outname, "w") as out:
        for chr in sort_seq_chr:
            # sort by start
            sort_seq_pos = sorted(chr_gene_pos[chr].items(), key=lambda d: int(d[1]), reverse=False)
            for i in sort_seq_pos:
                sort_seqname = i[0]
                seq = "".join(gene_seq[sort_seqname])
                strand = sort_seqname.split()[1].strip()
                for j in sort_seqname.split():
                    if "Len=" in j:
                        seqlen = j.split("=")[1]
                    if "Location" in j:
                        start = j.split(":")[1]
                        chr = j.split("=")[1].split(":")[0].strip()
                        end = int(int(start) + int(seqlen) - 1)
                    if "gene=" in j:
                        sim_seq_name = ">" + j.split("=")[1]
                out.write(sim_seq_name + "\t" + str(chr) + "\t" + str(start) + "\t" + str(end) + "\t" + strand + "\t" + str(seqlen) + "\n")
                out.write(seq + "\n")
    out.close()


def main():
    import optparse
    usage = """python -m plastidUtilis.Sort -f <Geseq_out_seq> -o <name_of_output> --header <Input_header>
                                                                            -Joe"""

    parser = optparse.OptionParser(usage)
    parser.add_option("-f", dest="Geseq_out", help="cds seqence extract from GeSeq output gff3",
                    metavar="FILE", action="store", type="string")
    parser.add_option("--header", dest="header", help="header of genome sequence, except the number",
                    metavar="STRING", action="store", type="string")
    parser.add_option("-o", dest="outname", help="output file name",
                    metavar="FILE", action="store", type="string")
    (options, args) = parser.parse_args()
    header_len = int(len(options.header))
    
    chr_gene_pos, gene_seq, chr_list = ParseFasta(options.Geseq_out)
    SortByChromosome(chr_gene_pos, gene_seq, chr_list, options.outname, header_len)


if __name__ == "__main__":
    main()


'''this module is able to execute local-installed BLAST to conduct sequence alignemnt analysis with preset parameters'''
import optparse
import os


def RunBLAST(INPUT, BLASTDATABASE, OUTPUT):
    trn_dict = {}
    # with open(tRNA_file) as f:
    with open(INPUT) as f:
        data = f.readlines()
        trn_dict["trash"] = []
        for line in data:
            if line.startswith(">"):
                seqname = line.strip().replace("\t", "@")[1:]
                if seqname not in trn_dict.keys():
                    trn_dict[seqname] = []
                elif seqname in trn_dict.keys():
                    seqname = "trash"
            else:
                trn_dict[seqname].append(line.strip())

    del trn_dict['trash']
    with open("temp_tRNA.fasta", "w") as t:
        for k, v in trn_dict.items():
            t.write(">" + k + "\n")
            t.write("".join(v) + "\n")
    t.close()

    print(len(trn_dict.keys()))
    os.system("makeblastdb -in %s -dbtype nucl -out %s" % (BLASTDATABASE, BLASTDATABASE))
    os.system(
        "blastn -query temp_tRNA.fasta -db %s -out blast.txt -perc_identity 90 -max_target_seqs 1 -subject_besthit  -num_threads 6 -outfmt  '6 qseqid qlen sseqid slen pident length score' " % database_file)
    os.system("rm temp_tRNA.fasta")

    blast_dict = {}
    with open("blast.txt") as b:
        data = b.readlines()
        for line in data:
            qseqid = line.split()[0]
            qlen = line.split()[1]
            sseqid = line.split()[2]
            slen = line.split()[3]
            pident = line.split()[4]
            length = line.split()[5]
            score = line.split()[6].strip()
            # length constrain
            if float(int(qlen) / int(slen)) > 0.9 and float(int(qlen) / int(slen)) < 1.1:
                # alignment length constrain
                if float(int(length) / int(slen)) > 0.9 and float(pident) > 0.90:
                    blast_dict[qseqid] = sseqid.split("-")[1] + "-" + sseqid.split("-")[2][:3]

    with open(OUTPUT, "w") as out:
        for k, v in blast_dict.items():
            old_id = k.split("@")[0]
            chr = k.split("@")[1]
            start = k.split("@")[2]
            end = k.split("@")[3]
            chain = k.split("@")[4]
            length = k.split("@")[5]
            new_id = v
            # intron
            if int(end) - int(start) + 1 == int(length):
                info = [new_id, chr, start, end, chain, length]
            else:
                info = [new_id, chr, start, end, chain, length, "intron"]
            out.write(">" + "\t".join(info) + "\n")
            out.write("".join(trn_dict[k]) + "\n")
    out.close()


def main():
    usage = """python -m plastidUtilis -d <tRNA_database> -t <sort_tRNA> -o <output.fasta>
                                                                                    --Joe"""
    parser = optparse.OptionParser(usage)
    parser.add_option("-d", dest="tRNA_database", help="tRNA blast database",
                    metavar="FILE", action="store", type="string")
    parser.add_option("-t", dest="sort_tRNA", help="sort_tRNA.fasta",
                    metavar="OUT", action="store", type="string")
    parser.add_option("-o", dest="output", help="output file name",
                    metavar="OUT", action="store", type="string")
    (options, args) = parser.parse_args()

    database_file = options.tRNA_database
    tRNA_file = options.sort_tRNA
    out_file = options.output

    RunBLAST(tRNA_file, database_file, out_file)


if __name__ == "__main__":
    main()


'''this module is used to translate the sorted CDS sequence into protein sequence
Note: this module now cannot tackle partial codon problems
# ---------------------------------------------------------------------------
BiopythonWarning: Partial codon, len(sequence) not a multiple of three. 
Explicitly trim the sequence or add trailing N before translation. 
This may become an error in future.
# ---------------------------------------------------------------------------
'''
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


def DNATranslate(inputname, outputname):
    '''
    Do DNA sequence translation all at once
    '''
    m_records = []
    # PROTEIN = open(output, 'w')
    with open(inputname, 'rt') as input:
        for record in SeqIO.parse(input, 'fasta'):
            header = record.description
            seq = record.seq.translate()
            
            # Direct write to output
            # PROTEIN.write(f'>{header}\n')
            # PROTEIN.write(f'{seq}\n')
            
            # Save as SeqRecords
            m_records.append(SeqRecord(
                Seq(seq),
                id=header,
                description=''
                ))
    
    SeqIO.write(m_records, outputname, "fasta")

def main():
    import optparse

    usage = """python -m plastidUtilis.Translate -i <sorted_CDS> -o <output_protein_sequence>
                                                                                -- Youpu=Chen"""
    parser = optparse.OptionParser(usage)
    parser.add_option("-i", dest="input", help="input sorted CDS fasta",
                    metavar="FILE", action="store", type="string")
    parser.add_option("-o", dest="output", help="output filename",
                    metavar="OUT", action="store", type="string")
    (options, args) = parser.parse_args()

    DNATranslate(options.input, options.output)


if __name__ == "__main__":
    main()





'''this module is used to conduct basic statistics based on multiple bio-formats'''
from Bio import SeqIO

def SeqLength(filename):
    '''Calculate sequence length in a given fasta file'''
    with open(filename) as input:
        for record in SeqIO.parse(input, 'fasta'):
            print(len(record.seq))

def main():
    import optparse

    usage = """python -m plastidUtilis.Stats -i <sorted_CDS>
                                                            -- Youpu=Chen"""
    parser = optparse.OptionParser(usage)
    parser.add_option("-i", dest="input", help="input fasta",
                    metavar="FILE", action="store", type="string")
    (options, args) = parser.parse_args()

    SeqLength(options.input)


if __name__ == "__main__":
    main()